const WebSocket = require('ws');
const AWS = require('aws-sdk');

const timestreamwrite = new AWS.TimestreamWrite({apiVersion: '2018-11-01'});
 
const ws = new WebSocket(`wss://ws.finnhub.io?token=${process.env.FINNHUB_TOKEN}`);

let buffer = [];
const symbols = {};

function flushInBackground() {
  if (buffer.length === 0) {
    return;
  }
  console.log('flushing...', buffer.length);
  const records = [];
  buffer.forEach(trade => {
    let microsecondOffset = 0;
    if (trade.s in symbols) {
      const symbol = symbols[trade.s];
      if (symbol.t === trade.t) {
        microsecondOffset = symbol.microsecondOffset + 1;
      }
    }
    symbols[trade.s] = {
      t: trade.t,
      microsecondOffset
    };
    if (microsecondOffset > 999) {
      console.error(`microsecondOffset > 999 for symbol ${trade.s} at time ${trade.t}, skip trade`);
      return;
    }
    const time = `${trade.t}${microsecondOffset.toString().padStart(3, '0')}`;
    records.push({
      Dimensions: [{
        Name: 'symbol',
        Value: trade.s,
        DimensionValueType: 'VARCHAR'
      }],
      MeasureName: 'price',
      MeasureValue: `${trade.p}`,
      MeasureValueType: 'DOUBLE',
      Time: time,
      TimeUnit: 'MICROSECONDS'
    });
    records.push({
      Dimensions: [{
        Name: 'symbol',
        Value: trade.s,
        DimensionValueType: 'VARCHAR'
      }],
      MeasureName: 'size',
      MeasureValue: `${trade.v}`,
      MeasureValueType: 'BIGINT',
      Time: time,
      TimeUnit: 'MICROSECONDS'
    });
  });
  buffer = [];
  timestreamwrite.writeRecords({
    DatabaseName: 'usstocks',
    Records: records,
    TableName: 'trades',
  }, (err) => {
    if (err) {
      console.error('can not flush, skipped', records.length);
      console.error(err);
    } else {
      console.log('flushed', records.length);
    }
  });
}

setInterval(flushInBackground, 250); // flush at least every 250 ms

ws.on('open', () => {
  console.log('open');
  ws.send(JSON.stringify({type: 'subscribe', symbol: 'AMZN'})); // Amazon
  ws.send(JSON.stringify({type: 'subscribe', symbol: 'AAPL'})); // Apple
  ws.send(JSON.stringify({type: 'subscribe', symbol: 'MSFT'})); // Microsoft
});
 
ws.on('message', (data) => {
  const msg = JSON.parse(data);
  if (msg.type === 'trade') {
    msg.data.forEach(d => {
      buffer.push(d);
      if (buffer.length >= 50) {
        flushInBackground();
      }
    });
  }
});
